@@ -9,21 +9,43 @@ import blazeutils.strings
 import pytz
 import six
 import sqlalchemy as sa
+import wrapt
 
 import keg_elements.decorators as decor
 import keg_elements.db.columns as columns
 import keg_elements.db.utils as dbutils
+from keg_elements.extensions import lazy_gettext as _
 
 
 might_commit = decor.keyword_optional('_commit', after=dbutils.session_commit, when_missing=True)
 might_flush = decor.keyword_optional('_flush', after=dbutils.session_flush)
 
 
+@wrapt.decorator
+def kwargs_match_entity(wrapped, instance, args, kwargs):
+    """
+    Asserts that the kwargs passed to the wrapped method match the columns/relationships
+    of the entity.
+    """
+    if kwargs.get('_check_kwargs', True):
+        insp = sa.inspection.inspect(instance)
+
+        # Only allow kwargs that correspond to a column or relationship on the entity
+        allowed_keys = {col.key for col in insp.columns} | set(insp.relationships.keys())
+
+        # Ignore kwargs starting with "_"
+        kwarg_keys = set(key for key in kwargs if not key.startswith('_'))
+        extra_kwargs = kwarg_keys - allowed_keys
+        assert not extra_kwargs, _('Unknown column or relationship names in kwargs: {kwargs!r}',
+                                   kwargs=sorted(extra_kwargs))
+    return wrapped(*args, **kwargs)
+
+
 class DefaultColsMixin(object):
     id = sa.Column('id', sa.Integer, primary_key=True)
-    created_utc = sa.Column(ArrowType, nullable=False, default=arrow.now,
+    created_utc = sa.Column(ArrowType, nullable=False, default=arrow.utcnow,
                             server_default=dbutils.utcnow())
-    updated_utc = sa.Column(ArrowType, nullable=False, default=arrow.now, onupdate=arrow.now,
+    updated_utc = sa.Column(ArrowType, nullable=False, default=arrow.utcnow, onupdate=arrow.utcnow,
                             server_default=dbutils.utcnow())
 
 
@@ -69,7 +91,7 @@ class MethodsMixin:
                 update_related(key, value, prop)
             else:
                 raise NotImplementedError(
-                    'Updating {} property types is not implemented.')
+                    _('Updating property types is not implemented.'))
 
     def to_dict(self, exclude=frozenset(), hybrids=frozenset()):
         """Covert the object properties to a dictionary.
@@ -140,8 +162,8 @@ class MethodsMixin:
                                    for x in cls.primary_keys()
                                    if x is not None]
         except KeyError:
-            raise AttributeError('No primary key was found in `oid` or `kwargs`'
-                                 ' for which to retrieve the object to edit')
+            raise AttributeError(_('No primary key was found in `oid` or `kwargs`'
+                                   ' for which to retrieve the object to edit'))
 
         obj = cls.query.get(primary_keys)
         obj.from_dict(kwargs)
@@ -185,6 +207,7 @@ class MethodsMixin:
 
         return [items(obj) for obj in result]
 
+    @kwargs_match_entity
     @classmethod
     def testing_create(cls, **kwargs):
         """Create an object for testing with default data appropriate for the field type
@@ -199,7 +222,7 @@ class MethodsMixin:
                                  controlled by this setting.
         """
 
-        NUMERIC_HIGH, NUMERIC_LOW = kwargs.pop('_numeric_defaults_range', (-100, 100))
+        numeric_range = kwargs.pop('_numeric_defaults_range', None)
 
         insp = sa.inspection.inspect(cls)
 
@@ -213,22 +236,36 @@ class MethodsMixin:
         for column in (col for col in insp.columns if not skippable(col)):
             try:
                 kwargs[column.key] = cls.random_data_for_column(
-                    column, NUMERIC_HIGH, NUMERIC_LOW)
+                    column, numeric_range)
             except ValueError:
                 pass
 
         return cls.add(**kwargs)
 
     @classmethod
-    def random_data_for_column(cls, column, numeric_high, numeric_low):
+    def random_data_for_column(cls, column, numeric_range):
+        if 'randomdata' in column.info:
+            if type(column.info['randomdata']) is str:
+                # assume randomdata the is name of a method on the class
+                callable = getattr(cls, column.info['randomdata'])
+                data = callable()
+                return data
+
+            return column.info['randomdata']()
+
+        default_range = (-100, 100) if numeric_range is None else numeric_range
         if isinstance(column.type, sa.types.Enum):
             return random.choice(column.type.enums)
         elif isinstance(column.type, sa.types.Boolean):
             return random.choice([True, False])
         elif isinstance(column.type, sa.types.Integer):
-            return random.randint(numeric_high, numeric_low)
+            return random.randint(*default_range)
+        elif isinstance(column.type, sa.types.Float):
+            return random.uniform(*default_range)
         elif isinstance(column.type, sa.types.Numeric):
-            return random.uniform(numeric_high, numeric_low)
+            if numeric_range is not None or column.type.scale is None:
+                return random.uniform(*default_range)
+            return dbutils.random_numeric(column)
         elif isinstance(column.type, sa.types.Date):
             return dt.date.today()
         elif isinstance(column.type, sa.types.DateTime):
@@ -241,7 +278,7 @@ class MethodsMixin:
             return random.choice(pytz.common_timezones)
         elif isinstance(column.type, (sa.types.String, sa.types.Unicode)):
             return blazeutils.strings.randchars(min(column.type.length or 25, 25))
-        raise ValueError('No randomization for this column type')
+        raise ValueError(_('No randomization for this column type'))
 
     @might_commit
     @might_flush
@@ -268,3 +305,87 @@ class MethodsMixin:
 
 class DefaultMixin(DefaultColsMixin, MethodsMixin):
     pass
+
+
+class HardDeleteProhibitedError(Exception):
+    pass
+
+
+class SoftDeleteMixin:
+    """SoftDeleteMixin alters the way deletes are performed by adding a deleted_utc column
+
+    A soft-delete doesn't actually remove the row, instead, it adds a time to the ``deleted_utc``
+    column which indicates when it was deleted.
+
+    .. note (NZ): This can complicate unique constraints, joins, and general business logic, but is
+    rather useful when you can't delete an object outright because it is connected to other
+    permanent objects which should never be deleted.
+
+    .. note (NZ): ``SoftDeleteMixin`` should appear before ``MethodsMixin`` as the parent class of
+    the created entity so that the ``delete`` and ``testing_create`` methods are called in the
+    correct order.
+
+        class MyTable(SoftDeleteMixin, MethodsMixin, Model):
+            column1 = sa.Column(sa.Numeric)
+
+
+    .. note (NZ): ``testing_create`` takes a special ``_is_deleted`` flag which enables you to
+    create an already deleted record or you can pass ``deleted_utc`` manually.
+
+    .. note (NZ): There is a new event registered within SQLAlchemy to prevent accidental deletions
+    of entities that inherit from this class.
+
+    This event is not fired for bulk operations (``session.delete``) and makes it possible to delete
+    everything.
+
+    To have finer grained control of this event for every entity, you can override the behavior by
+    implementing ``before_delete_event`` on the entity.
+
+        class MyEntity(SoftDeleteMixin, Model):
+            column1 = sa.Column(sa.Numeric)
+
+            # `mapper` and `connection` are passed through from the parent handler
+            def before_delete_event(self, mapper, connection):
+                # .. custom logic
+    """
+
+    deleted_utc = sa.Column(ArrowType, nullable=True)
+
+    @might_commit
+    @might_flush
+    @classmethod
+    def delete(cls, oid):
+        """Add the deleted_utc timestamp
+
+        :param oid: the object identifier, normally the primary key
+        :rtype: bool
+        :return: The result of the operation
+        """
+        return cls.query.filter(cls.id == oid).update(
+            {"deleted_utc": arrow.utcnow()},
+            synchronize_session=False
+        ) > 0
+
+    @might_commit
+    @might_flush
+    @classmethod
+    def testing_create(cls, *args, _is_deleted=False, **kwargs):
+        kwargs.setdefault('deleted_utc', arrow.utcnow() if _is_deleted else None)
+        return super().testing_create(*args, **kwargs)
+
+    @staticmethod
+    def sqla_before_delete_event(mapper, connection, target):
+        if hasattr(target, 'before_delete_event'):
+            target.before_delete_event(mapper, connection)
+        else:
+            raise HardDeleteProhibitedError(
+                'Unable to completely delete {}, this object implements soft-deletes.'.format(target))  # noqa
+
+    @might_commit
+    @classmethod
+    def delete_cascaded(cls):
+        cls.query.delete(synchronize_session=False)
+
+
+sa.event.listen(SoftDeleteMixin, 'before_delete', SoftDeleteMixin.sqla_before_delete_event,
+                propagate=True)
